using System.Collections;
using System.Collections.Generic;

using Syntriax.Engine.Core;

namespace Syntriax.Engine.Systems.Tween;

public class TweenManager : Behaviour, ITweenManager
{
    private CoroutineManager coroutineManager = null!;

    private readonly Dictionary<ITween, IEnumerator> runningCoroutines = [];
    private readonly Queue<Tween> pool = new();

    public ITween StartTween(float duration, ITweenManager.TweenSetCallback? setCallback = null)
    {
        Tween tween = Get(duration);
        tween.OnUpdated.AddListener(tween => setCallback?.Invoke(tween.Value));
        runningCoroutines.Add(tween, coroutineManager.StartCoroutine(RunTween(tween)));
        return tween;
    }

    private Tween Get(float duration)
    {
        if (!pool.TryDequeue(out Tween? result))
            result = new(duration);

        result.Duration = duration;
        return result;
    }

    private void Return(Tween tween)
    {
        if (pool.Contains(tween))
            return;

        tween.Wipe();
        pool.Enqueue(tween);
    }

    private IEnumerator RunTween(Tween tween)
    {
        tween.State = TweenState.Playing;
        yield return null;

        while (true)
        {
            if (tween.State.CheckFlag(TweenState.Cancelled | TweenState.Completed))
                break;

            if (tween.State.CheckFlag(TweenState.Paused))
            {
                yield return null;
                continue;
            }

            tween.Counter += Universe.Time.DeltaTime;
            yield return null;
        }

        runningCoroutines.Remove(tween);
        Return(tween);
    }

    public void CancelTween(ITween tween)
    {
        if (!runningCoroutines.TryGetValue(tween, out IEnumerator? runningCoroutine))
            return;

        tween.State = TweenState.Cancelled;
        coroutineManager.StopCoroutine(runningCoroutine);
        runningCoroutines.Remove(tween);
        Return((Tween)tween);
    }

    protected override void OnEnteredUniverse(IUniverse universe)
    {
        coroutineManager = universe.FindRequiredBehaviour<CoroutineManager>();
    }

    protected override void OnExitedUniverse(IUniverse universe)
    {
        coroutineManager = null!;
    }
}
